import { Expose, Transform, plainToClass, classToPlain } from 'class-transformer';
import { getClassValidatorMessagesStorage, Validator, getMetadataStorage, ValidationTypes, validateSync } from 'class-validator-multi-lang';
import { BehaviorSubject, Observable, Subject, combineLatest } from 'rxjs';
import { FormControl, FormGroup, FormArray, FormBuilder } from '@angular/forms';
import 'reflect-metadata';
import stringify from 'fast-safe-stringify';
import { distinctUntilChanged, tap, take } from 'rxjs/operators';
import stringHash from 'string-hash';

/**
 * @fileoverview added by tsickle
 * Generated from: lib/constants/default.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const DEFAULT_CLASS_TRANSFORM_OPTIONS = {
    strategy: 'excludeAll',
};
/** @type {?} */
const DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS = {
    strategy: 'exposeAll',
};
/** @type {?} */
const DEFAULT_CLASS_VALIDATOR_OPTIONS = { validationError: { target: false }, always: true };

/**
 * @fileoverview added by tsickle
 * Generated from: lib/decorators/expose-nested.decorator.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// for correct work expose with js object
// https://github.com/typestack/class-transformer/issues/365
/**
 * @param {?=} options
 * @return {?}
 */
function ExposeNested(options) {
    /** @type {?} */
    const exposeFn = Expose(options);
    /** @type {?} */
    const transformFn = (/**
     * @param {?} propertyKey
     * @return {?}
     */
    (propertyKey) => Transform((/**
     * @param {?} params
     * @return {?}
     */
    (params) => {
        return ((params.obj && Object.getOwnPropertyDescriptor(params.obj, propertyKey)
            ? params.obj[propertyKey]
            : options === null || options === void 0 ? void 0 : options.default) || params.value);
    })));
    return (/**
     * @param {?} target
     * @param {?} propertyKey
     * @return {?}
     */
    (target, propertyKey) => {
        transformFn(propertyKey)(target, propertyKey);
        exposeFn(target, propertyKey);
    });
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/get-global.util.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * This function returns the global object across Node and browsers.
 *
 * Note: `globalThis` is the standardized approach however it has been added to
 * Node.js in version 12. We need to include this snippet until Node 12 EOL.
 * @return {?}
 */
function getGlobal() {
    if (typeof globalThis !== 'undefined') {
        return globalThis;
    }
    if (typeof global !== 'undefined') {
        return global;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'window'.
    if (typeof window !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'window'.
        return window;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'self'.
    if (typeof self !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'self'.
        return self;
    }
    throw Error('Can not work with the global object');
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/storages/class-validator-messages.storage.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const CLASS_VALIDATOR_MESSAGES_STORAGE = 'CLASS_VALIDATOR_MESSAGES_STORAGE';
/**
 * @return {?}
 */
function getValidatorMessagesStorage() {
    const global = getGlobal();
    if (!global[CLASS_VALIDATOR_MESSAGES_STORAGE]) {
        global[CLASS_VALIDATOR_MESSAGES_STORAGE] = new BehaviorSubject(getClassValidatorMessagesStorage());
    }
    return global[CLASS_VALIDATOR_MESSAGES_STORAGE];
}
/**
 * @param {?} messages
 * @return {?}
 */
function updateValidatorMessagesStorage(messages) {
    getValidatorMessagesStorage().next(messages);
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/storages/class-validator-titles.storage.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const CLASS_VALIDATOR_TITLES_STORAGE = 'CLASS_VALIDATOR_TITLES_STORAGE';
/**
 * @return {?}
 */
function getValidatorTitlesStorage() {
    const global = getGlobal();
    if (!global[CLASS_VALIDATOR_TITLES_STORAGE]) {
        global[CLASS_VALIDATOR_TITLES_STORAGE] = new BehaviorSubject({});
    }
    return global[CLASS_VALIDATOR_TITLES_STORAGE];
}
/**
 * @param {?} titles
 * @return {?}
 */
function updateValidatorTitlesStorage(titles) {
    getValidatorTitlesStorage().next(titles);
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/dictionary.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 * @template T
 */
function Dictionary() { }

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/dynamic-form-builder-options.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function DynamicFormBuilderOptions() { }
if (false) {
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.factoryFormBuilder;
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.factoryDynamicFormGroup;
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.classValidatorOptions;
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.classTransformOptions;
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.classTransformToPlainOptions;
    /** @type {?|undefined} */
    DynamicFormBuilderOptions.prototype.validateAllFormFields;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/dynamic-form-group-config.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function DynamicFormGroupConfig() { }
if (false) {
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.validator;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.asyncValidator;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.validators;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.asyncValidators;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.updateOn;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.classValidatorOptions;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.classTransformOptions;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.classTransformToPlainOptions;
    /** @type {?|undefined} */
    DynamicFormGroupConfig.prototype.validateAllFormFields;
}
/**
 * @param {?} options
 * @return {?}
 */
function isDynamicFormGroupConfig(options) {
    return (options &&
        (Object.getOwnPropertyDescriptor(options, 'classValidatorOptions') ||
            Object.getOwnPropertyDescriptor(options, 'classTransformOptions') ||
            Object.getOwnPropertyDescriptor(options, 'classTransformToPlainOptions') ||
            Object.getOwnPropertyDescriptor(options, 'validateAllFormFields')));
}
/**
 * @param {?} options
 * @return {?}
 */
function isLegacyOrOpts(options) {
    return options && (!!options['validator'] || !!options['asyncValidator']);
}
/**
 * @param {?} options
 * @return {?}
 */
function isAbstractControlOptions(options) {
    return options && (!!options.validators || !!options.asyncValidators || !!options.updateOn);
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/dynamic-form-group-field.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function DynamicFormGroupField() { }
if (false) {
    /** @type {?} */
    DynamicFormGroupField.prototype.data;
    /** @type {?} */
    DynamicFormGroupField.prototype.validationFunctions;
    /** @type {?} */
    DynamicFormGroupField.prototype.validationDefinitions;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/error-property-name.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/form-model.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/short-validation-errors.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function ShortValidationErrors() { }

/**
 * @fileoverview added by tsickle
 * Generated from: lib/types/validator-function-type.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @record
 */
function ValidatorFunctionSyncType() { }
if (false) {
    /** @type {?} */
    ValidatorFunctionSyncType.prototype.type;
    /** @type {?} */
    ValidatorFunctionSyncType.prototype.validator;
}
/**
 * @record
 */
function ValidatorFunctionAsyncType() { }
if (false) {
    /** @type {?} */
    ValidatorFunctionAsyncType.prototype.type;
    /** @type {?} */
    ValidatorFunctionAsyncType.prototype.validator;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/validators/forever-invalid.validator.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} c
 * @return {?}
 */
function foreverInvalid(c) {
    return {
        foreverInvalid: {
            valid: false,
        },
    };
}
/** @type {?} */
const FOREVER_INVALID_NAME = 'foreverInvalid';

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/dynamic-form-control.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class DynamicFormControl extends FormControl {
    /**
     * @param {?} name
     * @param {?} fieldDefinition
     */
    constructor(name, fieldDefinition) {
        super(fieldDefinition.data instanceof Observable ? null : fieldDefinition.data, fieldDefinition.validationFunctions
            .filter((/**
         * @param {?} func
         * @return {?}
         */
        (func) => func.type === 'sync'))
            .map((/**
         * @param {?} func
         * @return {?}
         */
        (func) => (/** @type {?} */ (func.validator)))), fieldDefinition.validationFunctions
            .filter((/**
         * @param {?} func
         * @return {?}
         */
        (func) => func.type === 'async'))
            .map((/**
         * @param {?} func
         * @return {?}
         */
        (func) => (/** @type {?} */ (func.validator)))));
        this.controlName = name;
        this.validationDefinitions = fieldDefinition.validationDefinitions;
    }
}
if (false) {
    /** @type {?} */
    DynamicFormControl.prototype.controlName;
    /** @type {?} */
    DynamicFormControl.prototype.validationDefinitions;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/dynamic-form-group.util.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const cloneDeep = require('lodash.clonedeep');
/** @type {?} */
const mergeWith = require('lodash.mergewith');
/**
 * @param {?} errors
 * @return {?}
 */
function transformValidationErrors(errors) {
    /** @type {?} */
    const customErrors = {};
    errors.forEach((/**
     * @param {?} error
     * @return {?}
     */
    (error) => {
        if (error && error.constraints !== undefined) {
            Object.keys(error.constraints).forEach((/**
             * @param {?} key
             * @return {?}
             */
            (key) => {
                if (!customErrors[error.property]) {
                    customErrors[error.property] = [];
                }
                if (error.constraints && ((/** @type {?} */ (customErrors[error.property]))).indexOf(error.constraints[key]) === -1) {
                    ((/** @type {?} */ (customErrors[error.property]))).push(error.constraints[key]);
                }
            }));
        }
        if (error.children !== undefined && error.children.length) {
            customErrors[error.property] = transformValidationErrors(error.children);
        }
    }));
    return customErrors;
}
/**
 * @param {?=} externalErrors
 * @param {?=} validationErrors
 * @return {?}
 */
function mergeErrors(externalErrors, validationErrors) {
    /** @type {?} */
    const clonedExternalErrors = cloneDeep(externalErrors);
    return mergeWith(clonedExternalErrors, validationErrors, (/**
     * @param {?} objValue
     * @param {?} srcValue
     * @return {?}
     */
    (objValue, srcValue) => {
        if (canMerge()) {
            return objValue.concat(srcValue);
        }
        /**
         * @return {?}
         */
        function canMerge() {
            return (Array.isArray(objValue) && Array.isArray(srcValue) && !objValue.find((/**
             * @param {?} objItem
             * @return {?}
             */
            (objItem) => srcValue.includes(objItem))));
        }
    }));
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/get-or-set-empty-object.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} factoryModel
 * @return {?}
 */
function getOrSetEmptyObject(factoryModel) {
    if (!factoryModel) {
        return undefined;
    }
    if (!factoryModel.emptyObject) {
        factoryModel.emptyObject = new factoryModel();
    }
    return factoryModel.emptyObject;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/has-to-json.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} object
 * @return {?}
 */
function hasToJSON(object) {
    return object ? typeof object.toJSON === 'function' : false;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/dynamic-form-group.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const cloneDeep$1 = require('lodash.clonedeep');
/** @type {?} */
const validator = new Validator();
/**
 * @template TModel
 */
class DynamicFormGroup extends FormGroup {
    /**
     * @param {?} factoryModel
     * @param {?=} fields
     * @param {?=} _validatorOrOpts
     * @param {?=} _asyncValidator
     * @param {?=} options
     */
    constructor(factoryModel, fields, _validatorOrOpts, _asyncValidator, options) {
        super({}, _validatorOrOpts, _asyncValidator);
        this.factoryModel = factoryModel;
        this.fields = fields;
        this._validatorOrOpts = _validatorOrOpts;
        this._asyncValidator = _asyncValidator;
        this.options = options;
        this.nativeValidateErrors = new BehaviorSubject({});
        this.customValidateErrors = new BehaviorSubject({});
        this.objectChange = new Subject();
        /*
        const classValidators = DynamicFormGroup.getClassValidators<TModel>(
          this.factoryModel,
          this.fields,
          this.defaultValidatorOptions
        );
        const formGroup = this._fb.group(
          classValidators
        );
        Object.keys(formGroup.controls).forEach(key => {
          this.addControl(key, formGroup.controls[key]);
        });
        this.valueChanges.subscribe(data => {
          this.validate(
            undefined,
            this.defaultValidatorOptions
          );
        });*/
        if (options === null || options === void 0 ? void 0 : options.classTransformOptions) {
            this._classTransformOptions = options.classTransformOptions;
        }
        if (options === null || options === void 0 ? void 0 : options.classTransformToPlainOptions) {
            this._classTransformToPlainOptions = options.classTransformToPlainOptions;
        }
        if (options === null || options === void 0 ? void 0 : options.classValidatorOptions) {
            this._validatorOptions = options.classValidatorOptions;
        }
        if ((options === null || options === void 0 ? void 0 : options.validateAllFormFields) !== undefined) {
            this._validateAllFormFields = options.validateAllFormFields;
        }
        this.formFields = this.onlyFields(fields);
    }
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    subscribeToValueChanges(externalErrors, validatorOptions) {
        if (!this.valueChangesSubscription) {
            this.valueChangesSubscription = this.valueChanges
                .pipe(distinctUntilChanged(), tap((/**
             * @return {?}
             */
            () => this.validate(externalErrors, validatorOptions))))
                .subscribe({
                error: (/**
                 * @param {?} err
                 * @return {?}
                 */
                (err) => {
                    throw err;
                }),
            });
        }
    }
    // Getters & Setters
    /**
     * @param {?} externalErrors
     * @return {?}
     */
    set externalErrors(externalErrors) {
        this._externalErrors = externalErrors;
        this.validate();
    }
    /**
     * @return {?}
     */
    get externalErrors() {
        return this._externalErrors;
    }
    /**
     * @param {?} classTransformOptions
     * @return {?}
     */
    set classTransformOptions(classTransformOptions) {
        this._classTransformOptions = classTransformOptions;
        this.validate();
    }
    /**
     * @return {?}
     */
    get classTransformOptions() {
        return this._classTransformOptions;
    }
    /**
     * @param {?} classTransformToPlainOptions
     * @return {?}
     */
    set classTransformToPlainOptions(classTransformToPlainOptions) {
        this._classTransformToPlainOptions = classTransformToPlainOptions;
        this.validate();
    }
    /**
     * @return {?}
     */
    get classTransformToPlainOptions() {
        return this._classTransformToPlainOptions;
    }
    /**
     * @param {?} validatorOptions
     * @return {?}
     */
    set validatorOptions(validatorOptions) {
        this._validatorOptions = validatorOptions;
        this.validate();
    }
    /**
     * @return {?}
     */
    get validatorOptions() {
        return this._validatorOptions;
    }
    /**
     * @param {?} object
     * @return {?}
     */
    set object(object) {
        this.setObject(object);
    }
    /**
     * @return {?}
     */
    get object() {
        return this.getObject();
    }
    /**
     * @param {?} jsonData
     * @return {?}
     */
    set json(jsonData) {
        this.setJSON(jsonData);
    }
    /**
     * @return {?}
     */
    get json() {
        return this.getJSON();
    }
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validateStream(externalErrors, validatorOptions) {
        /** @type {?} */
        const argumentExternalErrors = externalErrors ? cloneDeep$1(externalErrors) : externalErrors;
        /** @type {?} */
        const argumentValidatorOptions = validatorOptions
            ? cloneDeep$1(validatorOptions)
            : validatorOptions;
        return combineLatest([getValidatorMessagesStorage(), getValidatorTitlesStorage()]).pipe(tap((/**
         * @param {?} __0
         * @return {?}
         */
        ([messages, titles]) => {
            validatorOptions = cloneDeep$1(this._validatorOptions || argumentValidatorOptions);
            externalErrors = cloneDeep$1(this._externalErrors || argumentExternalErrors);
            if (!validatorOptions) {
                validatorOptions = {};
            }
            if (!validatorOptions.messages) {
                validatorOptions.messages = {};
            }
            if (!validatorOptions.titles) {
                validatorOptions.titles = {};
            }
            if (messages) {
                validatorOptions.messages = Object.assign(Object.assign({}, cloneDeep$1(messages)), validatorOptions.messages);
            }
            if (titles) {
                validatorOptions.titles = Object.assign(Object.assign({}, cloneDeep$1(titles)), validatorOptions.titles);
            }
            /** @type {?} */
            const dataToValidate = this.object;
            /** @type {?} */
            const validationErrors = getValidateErrors(this, dataToValidate, validatorOptions);
            if (!externalErrors) {
                externalErrors = {};
            }
            /** @type {?} */
            const allErrors = mergeErrors(externalErrors, validationErrors);
            /** @type {?} */
            const allErrorsKeys = Object.keys(allErrors);
            this.markAsInvalidForExternalErrors(externalErrors);
            this.setCustomErrors(allErrors);
            // todo: refactor, invalidate form if exists any allErrors
            /** @type {?} */
            let usedForeverInvalid = false;
            if (!allErrorsKeys.find((/**
             * @param {?} key
             * @return {?}
             */
            (key) => key !== FOREVER_INVALID_NAME)) && this.get(FOREVER_INVALID_NAME)) {
                this.removeControl(FOREVER_INVALID_NAME);
                usedForeverInvalid = true;
            }
            if (this.valid && allErrorsKeys.length > 0 && !this.get(FOREVER_INVALID_NAME)) {
                this.addControl(FOREVER_INVALID_NAME, new FormControl('', [foreverInvalid]));
                usedForeverInvalid = true;
            }
            if (usedForeverInvalid) {
                this.updateValueAndValidity({
                    onlySelf: true,
                    emitEvent: false,
                });
            }
            if (this._validateAllFormFields) {
                this.validateAllFormFields();
            }
        })));
    }
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validateAsync(externalErrors, validatorOptions) {
        return this.validateStream(externalErrors, validatorOptions).pipe(take(1)).toPromise();
    }
    // Public API
    /**
     * @param {?=} externalErrors
     * @param {?=} validatorOptions
     * @return {?}
     */
    validate(externalErrors, validatorOptions) {
        if (this._validateSubscription) {
            this._validateSubscription.unsubscribe();
            this._validateSubscription = undefined;
        }
        this._validateSubscription = this.validateStream(externalErrors, validatorOptions).subscribe({
            error: (/**
             * @param {?} err
             * @return {?}
             */
            (err) => {
                throw err;
            }),
        });
    }
    /**
     * @param {?} allErrors
     * @param {?=} force
     * @return {?}
     */
    setCustomErrors(allErrors, force = false) {
        if (force || stringify(allErrors) !== stringify(this.customValidateErrors.value)) {
            this.formErrors = allErrors;
            this.customValidateErrors.next(this.formErrors);
        }
        /** @type {?} */
        const nativeValidateErrors = this.collectErrors(this);
        if (force || stringify(nativeValidateErrors) !== stringify(this.nativeValidateErrors.value)) {
            this.nativeValidateErrors.next(this.collectErrors(this));
        }
    }
    /**
     * @protected
     * @param {?} control
     * @param {?=} isRoot
     * @return {?}
     */
    collectErrors(control, isRoot = true) {
        if (control.controls) {
            return Object.assign(Object.assign({}, (isRoot ? this.errors : {})), Object.entries(control.controls).reduce((/**
             * @param {?} acc
             * @param {?} __1
             * @return {?}
             */
            (acc, [key, childControl]) => {
                /** @type {?} */
                const childErrors = this.collectErrors(childControl, false);
                if (childErrors && Object.keys(childErrors).length > 0) {
                    acc = Object.assign(Object.assign({}, acc), { [key]: Object.assign(Object.assign({}, (acc && acc[key] ? acc[key] : {})), childErrors) });
                }
                return acc;
            }), {}));
        }
        else {
            return control.errors;
        }
    }
    /**
     * @return {?}
     */
    validateAllFormFields() {
        /** @type {?} */
        let control;
        /** @type {?} */
        let formArrayControlsLength;
        Object.keys(this.controls).forEach((/**
         * @param {?} field
         * @return {?}
         */
        (field) => {
            control = this.get(field);
            // Control
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            // Group: recursive
            else if (control instanceof DynamicFormGroup) {
                control.validateAllFormFields();
            }
            // Array
            else if (control instanceof FormArray) {
                formArrayControlsLength = ((/** @type {?} */ (control))).controls.length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (((/** @type {?} */ (control))).controls[i] instanceof FormControl) {
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).markAsTouched({ onlySelf: true });
                    }
                    // Group in Array: recursive
                    else if (((/** @type {?} */ (control))).controls[i] instanceof DynamicFormGroup) {
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).validateAllFormFields();
                    }
                }
            }
        }));
        formArrayControlsLength = null;
        control = null;
    }
    /**
     * @return {?}
     */
    resetValidateAllFormFields() {
        this.markAsInvalidForExternalErrors({});
        /** @type {?} */
        let control;
        Object.keys(this.controls).forEach((/**
         * @param {?} field
         * @return {?}
         */
        (field) => {
            control = this.get(field);
            // Control
            if (control instanceof FormControl) {
                control.setErrors(null, { emitEvent: false });
                control.markAsUntouched({ onlySelf: true });
                control.markAsPristine({ onlySelf: true });
            }
            // Group: recursive
            else if (control instanceof DynamicFormGroup) {
                control.resetValidateAllFormFields();
            }
            // Array
            else if (control instanceof FormArray) {
                /** @type {?} */
                const formArrayControlsLength = ((/** @type {?} */ (control))).controls.length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (((/** @type {?} */ (control))).controls[i] instanceof FormControl) {
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).setErrors(null, { emitEvent: false });
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).markAsUntouched({ onlySelf: true });
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).markAsPristine({ onlySelf: true });
                    }
                    // Group in Array: recursive
                    else if (((/** @type {?} */ (control))).controls[i] instanceof DynamicFormGroup) {
                        ((/** @type {?} */ (((/** @type {?} */ (control))).controls[i]))).resetValidateAllFormFields();
                    }
                }
            }
        }));
        control = null;
        this.setCustomErrors({});
    }
    /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    classToClass(object) {
        if (hasToJSON(object)) {
            return new ((/** @type {?} */ (object))).constructor(((/** @type {?} */ (object))).toJSON());
        }
        return cloneDeep$1(object);
    }
    /**
     * @template TClassModel, Object
     * @param {?} cls
     * @param {?} plain
     * @return {?}
     */
    plainToClass(cls, plain) {
        if (hasToJSON(getOrSetEmptyObject(cls))) {
            return new cls(plain);
        }
        if (Object.keys(plain || {}).length === 0) {
            return this.classToClass(getOrSetEmptyObject(cls));
        }
        if (plain instanceof cls) {
            return this.classToClass(plain);
        }
        if (stringify(plain) === stringify(getOrSetEmptyObject(cls))) {
            return this.classToClass(getOrSetEmptyObject(cls));
        }
        return plainToClass(cls, plain, Object.assign({}, this._classTransformOptions));
    }
    /**
     * @template TClassModel
     * @param {?} object
     * @return {?}
     */
    classToPlain(object) {
        if (hasToJSON(object)) {
            return ((/** @type {?} */ (object))).toJSON();
        }
        return classToPlain(object, Object.assign({}, this._classTransformToPlainOptions));
    }
    /**
     * @deprecated please use sync method setExternalErrors(errors)
     * @param {?} externalErrors
     * @return {?}
     */
    setExternalErrorsAsync(externalErrors) {
        this.setExternalErrors(externalErrors);
        return Promise.resolve();
    }
    /**
     * @param {?} externalErrors
     * @return {?}
     */
    setExternalErrors(externalErrors) {
        this._externalErrors = externalErrors;
        this.validate();
    }
    /**
     * @return {?}
     */
    getExternalErrors() {
        return this._externalErrors;
    }
    /**
     * @deprecated please use sync method setExternalErrors({})
     * @return {?}
     */
    clearExternalErrorsAsync() {
        this.setExternalErrors({});
        return Promise.resolve();
    }
    /**
     * @return {?}
     */
    clearExternalErrors() {
        this.setExternalErrors({});
    }
    /**
     * @param {?=} validatorOptions
     * @return {?}
     */
    setValidatorOptions(validatorOptions) {
        this._validatorOptions = (/** @type {?} */ (validatorOptions));
        this.validate();
    }
    /**
     * @return {?}
     */
    getValidatorOptions() {
        return this._validatorOptions;
    }
    // Helpers
    /**
     * @protected
     * @param {?=} fields
     * @return {?}
     */
    onlyFields(fields) {
        /** @type {?} */
        const newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach((/**
             * @param {?} key
             * @return {?}
             */
            (key) => {
                if (fields[key] instanceof DynamicFormGroup) {
                    // Group: recursive
                    newFields[key] = this.onlyFields(((/** @type {?} */ (fields[key]))).formFields);
                }
                else {
                    // Array
                    if (fields[key] instanceof FormArray) {
                        if (((/** @type {?} */ (fields[key]))).controls[0] instanceof DynamicFormGroup) {
                            // Group within Array: recursive
                            newFields[key] = this.onlyFields(((/** @type {?} */ (((/** @type {?} */ (fields[key]))).controls[0]))).formFields);
                        }
                        else {
                            // Control within Array
                            newFields[key] = ((/** @type {?} */ (fields[key]))).controls[0].value;
                        }
                    }
                    else {
                        // Handle Control
                        newFields[key] = fields[key];
                    }
                }
            }));
        }
        return newFields;
    }
    /**
     * @protected
     * @param {?} errors
     * @param {?=} controls
     * @return {?}
     */
    markAsInvalidForExternalErrors(errors, controls) {
        /** @type {?} */
        const currentControls = controls ? controls : this.controls;
        Object.keys(currentControls).forEach((/**
         * @param {?} field
         * @return {?}
         */
        (field) => {
            /** @type {?} */
            const control = currentControls[field];
            // Control
            if (control instanceof FormControl) {
                if (errors && errors[field]) {
                    control.setErrors({ externalError: true });
                }
                else {
                    if (control.errors && control.errors.externalError === true) {
                        control.setErrors(null);
                    }
                }
            }
            // Group
            else if (control instanceof DynamicFormGroup) {
                control.markAsInvalidForExternalErrors(errors && errors[field] ? ((/** @type {?} */ (errors[field]))) : {});
            }
            // Array
            else if (control instanceof FormArray) {
                /** @type {?} */
                const formArrayControlsLength = ((/** @type {?} */ (control))).length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (control[i] instanceof FormControl) {
                        if (errors && errors[i] && errors[i][field]) {
                            control[i].setErrors({ externalError: true });
                        }
                        else if (control[i].errors && control[i].errors.externalError === true) {
                            control[i].setErrors(null);
                        }
                    }
                    // Group in Array
                    else if (control[i] instanceof DynamicFormGroup) {
                        control[i].markAsInvalidForExternalErrors(errors && errors[i] && errors[i][field] ? ((/** @type {?} */ (errors[i][field]))) : {});
                    }
                }
            }
        }));
    }
    /**
     * Recursively gets all values from the form controls and all sub form group and array controls and returns it as
     * an object
     * @protected
     * @return {?}
     */
    getObject() {
        // Initialize the shape of the response
        /** @type {?} */
        const object = this.classToClass(this._object || getOrSetEmptyObject(this.factoryModel) || undefined);
        if (object !== undefined) {
            // Recursively get the value of all fields
            Object.keys(this.controls)
                .filter((/**
             * @param {?} name
             * @return {?}
             */
            (name) => name !== FOREVER_INVALID_NAME))
                .forEach((/**
             * @param {?} key
             * @return {?}
             */
            (key) => {
                // Handle Group
                if (this.controls[key] instanceof DynamicFormGroup) {
                    object[key] = ((/** @type {?} */ (this.controls[key]))).object;
                }
                // Handle Form Array
                else if (this.controls[key] instanceof FormArray) {
                    // Initialize value
                    object[key] = [];
                    /** @type {?} */
                    const formArrayControlsLength = ((/** @type {?} */ (this.controls[key]))).controls.length;
                    /** @type {?} */
                    let value;
                    for (let i = 0; i < formArrayControlsLength; i++) {
                        if (((/** @type {?} */ (this.controls[key]))).controls[i] instanceof DynamicFormGroup) {
                            // Recursively get object group
                            value = ((/** @type {?} */ (((/** @type {?} */ (this.controls[key]))).controls[i]))).object;
                        }
                        else {
                            value = ((/** @type {?} */ (this.controls[key]))).controls[i].value;
                        }
                        if (value && Object.keys(value).length > 0) {
                            object[key].push(value);
                        }
                    }
                    value = null;
                }
                // Handle Control
                else {
                    if (this.controls[key]) {
                        object[key] = this.controls[key].value;
                    }
                }
            }));
        }
        return this.factoryModel ? this.plainToClass(this.factoryModel, object) : (/** @type {?} */ (object));
        // todo: try change to return this.classToClass(object);
    }
    /**
     * Sets the value of every control on the form and recursively sets the values of the controls
     * on all sub form groups
     *
     * @protected
     * @param {?} object the data to assign to all controls of the form group and sub groups
     * @return {?}
     */
    setObject(object) {
        // console.time(String(object));
        if (object instanceof this.factoryModel) {
            this._object = this.classToClass(object); // Ensure correct type
        }
        else {
            this._object = this.plainToClass(this.factoryModel, object); // Convert to Model type
            /*
            todo: remove later if speed is normal
            let objectWithOnlyFormFields: any = {};
            // for speed up on work with short version of model
            Object.keys(this.formFields).forEach((key) => (objectWithOnlyFormFields[key] = object ? object[key] : undefined));
            this._object = this.plainToClass(this.factoryModel, objectWithOnlyFormFields); // Convert to Model type
            objectWithOnlyFormFields = null;
            */
        }
        /** @type {?} */
        let newObject;
        /** @type {?} */
        let newValue;
        // Recursively set the value of all fields
        Object.keys(this.controls)
            .filter((/**
         * @param {?} name
         * @return {?}
         */
        (name) => name !== FOREVER_INVALID_NAME))
            .forEach((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            // Handle Group
            if (this.controls[key] instanceof DynamicFormGroup) {
                if (((/** @type {?} */ (this.controls[key])))._object !== this._object[key]) {
                    ((/** @type {?} */ (this.controls[key]))).setObject(this._object ? this._object[key] : {});
                }
            }
            // Handle FormArray
            else if (this.controls[key] instanceof FormArray) {
                /** @type {?} */
                let objectArray = this._object ? this._object[key] : [];
                /** @type {?} */
                let formArray = (/** @type {?} */ (this.controls[key]));
                /** @type {?} */
                let isFormGroup = formArray.controls[0] instanceof DynamicFormGroup;
                /** @type {?} */
                let firstFormGroup = (/** @type {?} */ (formArray.controls[0]));
                /** @type {?} */
                let formControl = (/** @type {?} */ (formArray.controls[0]));
                // Clear FormArray while retaining the reference
                /** @type {?} */
                let objectArrayLength = objectArray.length;
                /** @type {?} */
                let dynamicFormGroup;
                /** @type {?} */
                let controlValue;
                /** @type {?} */
                let newFormControl;
                /** @type {?} */
                let classValidators;
                /** @type {?} */
                let formGroup;
                /** @type {?} */
                let formArrayLength = formArray.length;
                for (let i = 0; i < formArrayLength; i++) {
                    if (formArray.controls[i] instanceof DynamicFormGroup &&
                        ((/** @type {?} */ (formArray.controls[i]))).valueChangesSubscription) {
                        ((/** @type {?} */ (formArray.controls[i]))).valueChangesSubscription.unsubscribe();
                    }
                }
                while (formArrayLength !== 0) {
                    formArray.removeAt(0);
                    formArrayLength--;
                }
                formArrayLength = null;
                for (let i = 0; i < objectArrayLength; i++) {
                    if (isFormGroup) {
                        // Create FormGroup
                        dynamicFormGroup = this.dynamicFormBuilder.factoryDynamicFormGroup(firstFormGroup.factoryModel, firstFormGroup.formFields);
                        dynamicFormGroup.setParent(formArray);
                        formGroup = this.originalFormBuilder.group(getClassValidators(firstFormGroup.factoryModel, firstFormGroup.formFields));
                        // Add all controls to the form group
                        Object.keys(formGroup.controls).forEach((/**
                         * @param {?} ctrlKey
                         * @return {?}
                         */
                        (ctrlKey) => {
                            dynamicFormGroup.addControl(ctrlKey, formGroup.controls[ctrlKey]);
                        }));
                        formArray.controls.push(dynamicFormGroup);
                    }
                    else {
                        // Create control
                        controlValue = this._object && objectArray && objectArray[i] ? objectArray[i] : undefined;
                        newFormControl = new FormControl(controlValue, formControl ? formControl.validator : undefined);
                        newFormControl.setParent(formArray);
                        // Add the control to the FormArray
                        formArray.controls.push(newFormControl);
                    }
                }
                /** @type {?} */
                let arrayLength = formArray.length;
                for (let i = 0; i < arrayLength; i++) {
                    if (isFormGroup) {
                        // Add a value change listener to the group. on change, validate
                        if (formArray.controls[i] instanceof DynamicFormGroup) {
                            // Recusrively set the object value
                            ((/** @type {?} */ (formArray.controls[i]))).setObject(this._object && objectArray && objectArray[i] ? objectArray[i] : {});
                        }
                    }
                }
                for (let i = 0; i < arrayLength; i++) {
                    if (isFormGroup) {
                        // Add a value change listener to the group. on change, validate
                        if (formArray.controls[i] instanceof DynamicFormGroup) {
                            ((/** @type {?} */ (formArray.controls[i]))).subscribeToValueChanges(undefined, this._validatorOptions);
                        }
                    }
                }
                classValidators = null;
                formGroup = null;
                controlValue = null;
                newFormControl = null;
                dynamicFormGroup = null;
                objectArrayLength = null;
                arrayLength = null;
                objectArray = null;
                formArray = null;
                isFormGroup = null;
                firstFormGroup = null;
                formControl = null;
            }
            // Handle Control
            else {
                newObject = this._object ? this._object[key] : [];
                newValue = this._object && newObject ? newObject : undefined;
                if (this.controls[key] && this.controls[key].value !== newValue) {
                    // very slow operations :(
                    this.controls[key].setValue(newValue, { emitEvent: false, onlySelf: true });
                }
            }
        }));
        newValue = null;
        newObject = null;
        this.updateValueAndValidity();
        this.objectChange.next(this._object);
        if (this._validateAllFormFields) {
            this.validateAllFormFields();
        }
        // console.timeEnd(String(object));
    }
    /**
     * @protected
     * @return {?}
     */
    getJSON() {
        return this.classToPlain(this.getObject());
    }
    /**
     * @protected
     * @param {?} jsonData
     * @return {?}
     */
    setJSON(jsonData) {
        this.setObject(this.plainToClass(this.factoryModel, jsonData));
    }
}
if (false) {
    /** @type {?} */
    DynamicFormGroup.prototype.nativeValidateErrors;
    /** @type {?} */
    DynamicFormGroup.prototype.customValidateErrors;
    /** @type {?} */
    DynamicFormGroup.prototype.formErrors;
    /** @type {?} */
    DynamicFormGroup.prototype.formFields;
    /** @type {?} */
    DynamicFormGroup.prototype.objectChange;
    /** @type {?} */
    DynamicFormGroup.prototype.valueChangesSubscription;
    /** @type {?} */
    DynamicFormGroup.prototype.dynamicFormBuilder;
    /** @type {?} */
    DynamicFormGroup.prototype.originalFormBuilder;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._object;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._externalErrors;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._validatorOptions;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._classTransformOptions;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._classTransformToPlainOptions;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._validateSubscription;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._validateAllFormFields;
    /** @type {?} */
    DynamicFormGroup.prototype.factoryModel;
    /** @type {?} */
    DynamicFormGroup.prototype.fields;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._validatorOrOpts;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype._asyncValidator;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormGroup.prototype.options;
}
/**
 * @template TModel
 * @param {?} factoryModel
 * @param {?=} fields
 * @param {?=} validatorOptions
 * @return {?}
 */
function getClassValidators(factoryModel, fields, validatorOptions) {
    /** @type {?} */
    const groups = validatorOptions ? validatorOptions.groups : undefined;
    /** @type {?} */
    const strictGroups = (validatorOptions && validatorOptions.strictGroups) || false;
    /** @type {?} */
    const always = (validatorOptions && validatorOptions.always) || false;
    // console.time(String(factoryModel));
    // Get the validation rules from the object decorators
    /** @type {?} */
    let allValidationMetadatas = getMetadataStorage().getTargetValidationMetadatas(factoryModel, '', always, strictGroups);
    // Get the validation rules for the validation group: https://github.com/typestack/class-validator-multi-lang#validation-groups
    /** @type {?} */
    let validationGroupMetadatas = getMetadataStorage().getTargetValidationMetadatas(factoryModel, '', always, strictGroups, groups);
    /** @type {?} */
    const formGroupFields = {};
    // Loop through all fields in the form definition
    if (fields) {
        /** @type {?} */
        let conditionalValidations;
        /** @type {?} */
        let allNestedValidations;
        /** @type {?} */
        let nestedGroupValidations;
        /** @type {?} */
        let fieldDefinition;
        /** @type {?} */
        let dynamicValidate;
        /** @type {?} */
        let customValidation;
        /** @type {?} */
        let objectToValidate;
        /** @type {?} */
        let nestedValidate;
        Object.keys(fields)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        (key) => key.indexOf('__') !== 0))
            .forEach((/**
         * @param {?} fieldName
         * @return {?}
         */
        (fieldName) => {
            // Conditional Validation for the field
            conditionalValidations = [];
            validationGroupMetadatas.forEach((/**
             * @param {?} validationMetadata
             * @return {?}
             */
            (validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.conditional.type)) {
                    conditionalValidations.push(validationMetadata);
                }
            }));
            // All Nested Validation for the field
            allNestedValidations = [];
            allValidationMetadatas.forEach((/**
             * @param {?} validationMetadata
             * @return {?}
             */
            (validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.nested.type)) {
                    allNestedValidations.push(validationMetadata);
                }
            }));
            // Nested Validation for the field for the requested class-validator-multi-lang group
            nestedGroupValidations = [];
            validationGroupMetadatas.forEach((/**
             * @param {?} validationMetadata
             * @return {?}
             */
            (validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.nested.type)) {
                    nestedGroupValidations.push(validationMetadata);
                }
            }));
            fieldDefinition = {
                data: formGroupFields[fieldName] || undefined,
                validationFunctions: [],
                validationDefinitions: [],
            };
            if (fieldDefinition.data === undefined) {
                fieldDefinition.data = fields[fieldName];
            }
            // TRY LINK EXISTS NATIVE VALIDATIONS, UNSTABLE !!!
            if (Array.isArray(fieldDefinition.data) &&
                fieldDefinition.data.length > 1 &&
                fieldDefinition.data.find((/**
                 * @param {?} validationFunction
                 * @param {?} index
                 * @return {?}
                 */
                (validationFunction, index) => index > 0 && typeof validationFunction === 'function'))) {
                fieldDefinition.data
                    .filter((/**
                 * @param {?} validationFunction
                 * @param {?} index
                 * @return {?}
                 */
                (validationFunction, index) => index > 0 && typeof validationFunction === 'function'))
                    .forEach((/**
                 * @param {?} validationFunction
                 * @return {?}
                 */
                (validationFunction) => fieldDefinition.validationFunctions.push({ type: 'sync', validator: validationFunction })));
                fieldDefinition.data = fieldDefinition.data[0];
            }
            validationGroupMetadatas.forEach((/**
             * @param {?} validationMetadata
             * @return {?}
             */
            (validationMetadata) => {
                if (validationMetadata.propertyName === fieldName &&
                    validationMetadata.type !== ValidationKeys.conditional.type) {
                    // Add all validation to the field except the @NestedValidation definition as
                    // being part of the form would imply it is validated if any other rules are present
                    if (validationMetadata.type !== ValidationKeys.nested.type) {
                        fieldDefinition.validationDefinitions.push(validationMetadata);
                    }
                    Object.keys(ValidationTypes).forEach((/**
                     * @param {?} typeKey
                     * @return {?}
                     */
                    (typeKey) => {
                        if (ValidationTypes.hasOwnProperty(typeKey)) {
                            // Handle Nested Validation
                            if (checkWithAllNestedValidations(allNestedValidations, nestedGroupValidations, fieldName)) {
                                if (isNestedValidate(validationMetadata, typeKey)) {
                                    objectToValidate =
                                        fields[fieldName] instanceof DynamicFormGroup ? fields[fieldName].object : undefined;
                                    nestedValidate = createNestedValidate(fieldName, objectToValidate, validationMetadata);
                                    setFieldData(fieldName, fieldDefinition, nestedValidate);
                                    objectToValidate = null;
                                    nestedValidate = null;
                                }
                            }
                            // Handle Custom Validation
                            if (isCustomValidate(validationMetadata, typeKey)) {
                                customValidation = createCustomValidation(fieldName, validationMetadata);
                                setFieldData(fieldName, fieldDefinition, customValidation);
                                customValidation = null;
                            }
                            // Handle remaining validation
                            if (isDynamicValidate(validationMetadata, typeKey)) {
                                dynamicValidate = createDynamicValidate(fieldName, validationMetadata, conditionalValidations);
                                setFieldData(fieldName, fieldDefinition, dynamicValidate);
                                dynamicValidate = null;
                            }
                        }
                    }));
                }
            }));
            // Convert to a structure, angular understands
            if (fieldDefinition.data instanceof DynamicFormGroup || fieldDefinition.data instanceof FormArray) {
                formGroupFields[fieldName] = fieldDefinition.data;
            }
            else {
                formGroupFields[fieldName] = new DynamicFormControl(fieldName, fieldDefinition);
            }
        }));
        dynamicValidate = null;
        customValidation = null;
        objectToValidate = null;
        nestedValidate = null;
        conditionalValidations = null;
        allNestedValidations = null;
        nestedGroupValidations = null;
        fieldDefinition = null;
    }
    allValidationMetadatas = null;
    validationGroupMetadatas = null;
    // console.timeEnd(String(factoryModel));
    return formGroupFields;
    // ******************************************************************************************
    // Local Helper functions to help make the main code more readable
    //
    /**
     * @param {?} fieldName
     * @param {?} objectToValidate
     * @param {?} validationMetadata
     * @return {?}
     */
    function createNestedValidate(fieldName, objectToValidate, validationMetadata) {
        return {
            type: 'sync',
            validator: (/**
             * @param {?} control
             * @return {?}
             */
            function (control) {
                /** @type {?} */
                let parent = control.parent;
                /** @type {?} */
                let validateErrors = (parent &&
                    parent.value &&
                    getValidateErrors(parent, objectToValidate !== undefined ? objectToValidate : control.value, validatorOptions)) ||
                    {};
                /** @type {?} */
                let validateState = Object.keys(validateErrors || {}).length === 0;
                /** @type {?} */
                const result = getIsValidResult(validateState, validationMetadata, 'nestedValidate');
                parent = null;
                validateErrors = null;
                validateState = null;
                return result;
            }),
        };
    }
    /**
     * @param {?} fieldName
     * @param {?} validationMetadata
     * @param {?} conditionalValidations
     * @return {?}
     */
    function createDynamicValidate(fieldName, validationMetadata, conditionalValidations) {
        return {
            type: 'sync',
            validator: (/**
             * @param {?} control
             * @return {?}
             */
            function (control) {
                if (!control) {
                    return null;
                }
                /** @type {?} */
                let validateErrors = setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions);
                /** @type {?} */
                let validateState = getAllErrors(validateErrors, fieldName).length === 0;
                /** @type {?} */
                const result = getIsValidResult(validateState, validationMetadata, 'dynamicValidate');
                validateErrors = null;
                validateState = null;
                return result;
            }),
        };
    }
    /**
     * @param {?} fieldName
     * @param {?} validationMetadata
     * @return {?}
     */
    function createCustomValidation(fieldName, validationMetadata) {
        return {
            type: 'sync',
            validator: (/**
             * @param {?} control
             * @return {?}
             */
            function (control) {
                /** @type {?} */
                let validateErrors = setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions);
                /** @type {?} */
                let validateState = validateErrors && validateErrors[fieldName] ? false : true;
                /** @type {?} */
                const result = getIsValidResult(validateState, validationMetadata, 'customValidation');
                validateErrors = null;
                validateState = null;
                return result;
            }),
        };
    }
    /**
     * @param {?} allNestedValidations
     * @param {?} nestedValidations
     * @param {?} key
     * @return {?}
     */
    function checkWithAllNestedValidations(allNestedValidations, nestedValidations, key) {
        /** @type {?} */
        const allNestedValidationsLength = allNestedValidations.length;
        /** @type {?} */
        const nestedValidationsLength = nestedValidations.length;
        return (allNestedValidationsLength === nestedValidationsLength ||
            (((fields && fields[key] instanceof DynamicFormGroup) || (fields && fields[key] instanceof FormArray)) &&
                allNestedValidationsLength > 0 &&
                nestedValidationsLength === 0));
    }
    /**
     * @param {?} validationMetadata
     * @param {?} typeKey
     * @return {?}
     */
    function isDynamicValidate(validationMetadata, typeKey) {
        return validationMetadata.type === ValidationTypes[typeKey] && validator[validationMetadata.type] !== undefined;
    }
    /**
     * marked with \@Validate(...)
     * https://github.com/typestack/class-validator-multi-lang#custom-validation-classes
     * @param {?} validationMetadata
     * @param {?} typeKey
     * @return {?}
     */
    function isCustomValidate(validationMetadata, typeKey) {
        return (isNotPropertyValidation(validationMetadata, typeKey) &&
            validationMetadata.type === ValidationKeys.custom.type &&
            typeKey === ValidationKeys.custom.typeKey);
    }
    /**
     * marked with \@ValidateNested()
     * https://github.com/typestack/class-validator-multi-lang#validating-nested-objects
     * @param {?} validationMetadata
     * @param {?} typeKey
     * @return {?}
     */
    function isNestedValidate(validationMetadata, typeKey) {
        return (isNotPropertyValidation(validationMetadata, typeKey) &&
            validationMetadata.type === ValidationKeys.nested.type &&
            typeKey === ValidationKeys.nested.typeKey);
    }
    /**
     * @param {?} validationMetadata
     * @param {?} typeKey
     * @return {?}
     */
    function isNotPropertyValidation(validationMetadata, typeKey) {
        return validationMetadata.type === ValidationTypes[typeKey] && validator[validationMetadata.type] === undefined;
    }
    /**
     * @param {?} fieldName
     * @param {?} fieldDefinition
     * @param {?} validationFunction
     * @return {?}
     */
    function setFieldData(fieldName, fieldDefinition, validationFunction) {
        /* todo: maybe not need, if enable this code, experemental mode not work
        if (fields[fieldName] instanceof DynamicFormGroup) {
          fields[fieldName].object = fields[fieldName].fields;
        }*/
        // Fill field data if empty
        if (fieldDefinition.data === undefined) {
            fieldDefinition.data = fields && fields[fieldName];
        }
        fieldDefinition.validationFunctions.push(validationFunction);
    }
    /**
     * @param {?} validateErrors
     * @param {?} fieldName
     * @return {?}
     */
    function getAllErrors(validateErrors, fieldName) {
        return Object.keys(validateErrors).filter((/**
         * @param {?} key
         * @return {?}
         */
        (key) => 
        // Check for nested/child errors
        validateErrors[key][fieldName] || validateErrors[fieldName]));
    }
}
// ***************************************************************
// Global Helper functions
//
/**
 * @param {?} validationMetadata
 * @param {?} fieldName
 * @param {?} validationMetadataType
 * @return {?}
 */
function isPropertyValidatorOfType(validationMetadata, fieldName, validationMetadataType) {
    return validationMetadata.propertyName === fieldName && validationMetadata.type === validationMetadataType;
}
/**
 * @param {?} fieldName
 * @param {?} control
 * @param {?=} validatorOptions
 * @return {?}
 */
function setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions) {
    /** @type {?} */
    let parent = control.parent instanceof DynamicFormGroup
        ? ((/** @type {?} */ (control.parent)))
        : control.parent
            ? control.parent
            : null;
    /** @type {?} */
    let object = control.parent instanceof DynamicFormGroup
        ? ((/** @type {?} */ (control.parent))).object
        : control.parent
            ? control.parent.value
            : {};
    if (object) {
        object[fieldName] = control.value;
    }
    /** @type {?} */
    const result = (parent && parent.value && getValidateErrors(parent, object, validatorOptions)) || {};
    parent = null;
    object = null;
    return result;
}
/**
 * @template T
 * @param {?} control
 * @param {?} dataToValidate
 * @param {?=} validatorOptions
 * @return {?}
 */
function getValidateErrors(control, dataToValidate, validatorOptions) {
    /** @type {?} */
    let validateKey = stringHash(stringify({ dataToValidate, validatorOptions }));
    /** @type {?} */
    let ctrl = ((/** @type {?} */ (control))) || {};
    /** @type {?} */
    let validatedErrors;
    if (ctrl.__prevValidateKey === validateKey && ctrl.__prevValidateErrors !== undefined) {
        // console.log('CACHED', validateKey);
        validatedErrors = ctrl.__prevValidateErrors;
    }
    else {
        // console.log('NO CACHED', validateKey);
        validatedErrors = transformValidationErrors(validateSync(dataToValidate, validatorOptions));
        ctrl.__prevValidateErrors = validatedErrors;
        ctrl.__prevValidateKey = validateKey;
    }
    validateKey = null;
    ctrl = null;
    return validatedErrors;
}
/**
 * @param {?} isValid
 * @param {?} validationMetadata
 * @param {?} errorType
 * @return {?}
 */
function getIsValidResult(isValid, validationMetadata, errorType) {
    return isValid
        ? null
        : {
            [errorType]: {
                valid: false,
                type: validationMetadata.type,
            },
        };
}
/** @type {?} */
const ValidationKeys = {
    nested: {
        type: 'nestedValidation',
        typeKey: 'NESTED_VALIDATION',
    },
    conditional: {
        type: 'conditionalValidation',
    },
    custom: {
        type: 'customValidation',
        typeKey: 'CUSTOM_VALIDATION',
    },
};

/**
 * @fileoverview added by tsickle
 * Generated from: lib/utils/dynamic-form-builder.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const cloneDeep$2 = require('lodash.clonedeep');
class DynamicFormBuilder extends FormBuilder {
    /**
     * @param {?=} options
     */
    constructor(options) {
        super();
        this.options = options;
        // need for createEmptyObject
        this.emptyDynamicFormGroup = this.factoryDynamicFormGroup(Object);
    }
    // ******************
    // Public API
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} controlsConfig
     * @param {?=} options
     * @return {?}
     */
    group(factoryModel, controlsConfig, options) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        // console.time(factoryModel.toString());
        if (!controlsConfig && !options) {
            options = {};
        }
        // Process the group with the controlsConfig passed into extra instead. (What does this accomplish?)
        if (controlsConfig &&
            (isAbstractControlOptions(controlsConfig) ||
                isLegacyOrOpts(controlsConfig) ||
                isDynamicFormGroupConfig(controlsConfig)) &&
            Object.keys(options || {}).length === 0) {
            return this.group(factoryModel, undefined, controlsConfig);
        }
        /** @type {?} */
        let extra = (/** @type {?} */ (cloneDeep$2(options)));
        /** @type {?} */
        let validators = null;
        /** @type {?} */
        let asyncValidators = null;
        /** @type {?} */
        let updateOn;
        if (extra != null) {
            if (isAbstractControlOptions(extra)) {
                // `extra` are `AbstractControlOptions`
                validators = extra.validators != null ? extra.validators : null;
                asyncValidators = extra.asyncValidators != null ? extra.asyncValidators : null;
                updateOn = extra.updateOn != null ? extra.updateOn : undefined;
            }
            if (isLegacyOrOpts(extra)) {
                // `extra` are legacy form group options
                validators = validators || [];
                if (extra.validator != null) {
                    validators.push(extra.validator);
                }
                asyncValidators = asyncValidators || [];
                if (extra.asyncValidator != null) {
                    asyncValidators.push(extra.asyncValidator);
                }
            }
        }
        else {
            extra = {};
        }
        if (((_a = this.options) === null || _a === void 0 ? void 0 : _a.classValidatorOptions) && !extra.classValidatorOptions) {
            extra.classValidatorOptions = (_b = this.options) === null || _b === void 0 ? void 0 : _b.classValidatorOptions;
        }
        if (((_c = this.options) === null || _c === void 0 ? void 0 : _c.classTransformOptions) && !extra.classTransformOptions) {
            extra.classTransformOptions = (_d = this.options) === null || _d === void 0 ? void 0 : _d.classTransformOptions;
        }
        if (((_e = this.options) === null || _e === void 0 ? void 0 : _e.classTransformToPlainOptions) && !extra.classTransformToPlainOptions) {
            extra.classTransformToPlainOptions = (_f = this.options) === null || _f === void 0 ? void 0 : _f.classTransformToPlainOptions;
        }
        if (((_g = this.options) === null || _g === void 0 ? void 0 : _g.validateAllFormFields) !== undefined && extra.validateAllFormFields === undefined) {
            extra.validateAllFormFields = (_h = this.options) === null || _h === void 0 ? void 0 : _h.validateAllFormFields;
        }
        // Set default classValidatorOptions
        if (!extra.classValidatorOptions) {
            extra.classValidatorOptions = DEFAULT_CLASS_VALIDATOR_OPTIONS;
        }
        if (!extra.classTransformOptions) {
            extra.classTransformOptions = DEFAULT_CLASS_TRANSFORM_OPTIONS;
        }
        if (!extra.classTransformToPlainOptions) {
            extra.classTransformToPlainOptions = DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS;
        }
        if (extra.validateAllFormFields === undefined) {
            extra.validateAllFormFields = false;
        }
        /** @type {?} */
        let newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = (/** @type {?} */ (controlsConfig));
        }
        // experimental
        if (controlsConfig === undefined) {
            newControlsConfig = Object.assign({}, this.createEmptyObject(factoryModel, { __experimental__: true }));
            if (newControlsConfig !== undefined) {
                Object.keys(newControlsConfig).forEach((/**
                 * @param {?} key
                 * @return {?}
                 */
                (key) => {
                    if (canCreateGroup() && newControlsConfig) {
                        // recursively create a dynamic group for the nested object
                        newControlsConfig[key] = this.group(newControlsConfig[key].constructor, undefined, {
                            classValidatorOptions: extra.classValidatorOptions,
                            classTransformOptions: extra.classTransformOptions,
                            classTransformToPlainOptions: extra.classTransformToPlainOptions,
                            validateAllFormFields: extra.validateAllFormFields,
                            asyncValidators,
                            updateOn,
                            validators,
                        });
                    }
                    else {
                        if (canCreateArray() && newControlsConfig) {
                            if (newControlsConfig[key][0].constructor) {
                                // recursively create an array with a group
                                newControlsConfig[key] = super.array(newControlsConfig[key].map((/**
                                 * @param {?} newControlsConfigItem
                                 * @return {?}
                                 */
                                (newControlsConfigItem) => this.group(newControlsConfigItem.constructor, undefined, {
                                    classValidatorOptions: extra.classValidatorOptions,
                                    classTransformOptions: extra.classTransformOptions,
                                    classTransformToPlainOptions: extra.classTransformToPlainOptions,
                                    validateAllFormFields: extra.validateAllFormFields,
                                    asyncValidators,
                                    updateOn,
                                    validators,
                                }))));
                            }
                            else {
                                // Create an array of form controls
                                newControlsConfig[key] = super.array(newControlsConfig[key].map((/**
                                 * @param {?} newControlsConfigItem
                                 * @return {?}
                                 */
                                (newControlsConfigItem) => this.control(newControlsConfigItem))));
                            }
                        }
                    }
                    /**
                     * @return {?}
                     */
                    function canCreateGroup() {
                        /** @type {?} */
                        const candidate = newControlsConfig && newControlsConfig[key];
                        return (candidate &&
                            !Array.isArray(candidate) &&
                            candidate.constructor &&
                            typeof candidate === 'object' &&
                            (candidate.length === undefined ||
                                (candidate.length !== undefined && Object.keys(candidate).length === candidate.length)));
                    }
                    /**
                     * @return {?}
                     */
                    function canCreateArray() {
                        if (Array.isArray(newControlsConfig && newControlsConfig[key]) === false) {
                            return false;
                        }
                        /** @type {?} */
                        const candidate = newControlsConfig && newControlsConfig[key][0];
                        return (candidate.constructor &&
                            typeof candidate === 'object' &&
                            (candidate.length === undefined ||
                                (candidate.length !== undefined && Object.keys(candidate).length === candidate.length)));
                    }
                }));
            }
        }
        // Remove empty
        validators = validators && validators.filter((/**
         * @param {?} validator
         * @return {?}
         */
        (validator) => validator));
        asyncValidators = asyncValidators && asyncValidators.filter((/**
         * @param {?} validator
         * @return {?}
         */
        (validator) => validator));
        // Create an Angular group from the top-level object
        /** @type {?} */
        let classValidators = getClassValidators(factoryModel, newControlsConfig, extra && extra.classValidatorOptions);
        /** @type {?} */
        let formGroup = super.group(classValidators, Object.assign(Object.assign(Object.assign({}, (asyncValidators || {})), (updateOn || {})), (validators || {})));
        // Initialize the resulting group
        /** @type {?} */
        const dynamicFormGroup = this.factoryDynamicFormGroup(factoryModel, newControlsConfig, {
            asyncValidators,
            updateOn,
            validators,
        }, undefined, extra);
        // Add all angular controls to the resulting dynamic group
        Object.keys(formGroup.controls).forEach((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        }));
        // Add a listener to the dynamic group for value changes; on change, execute validation
        dynamicFormGroup.subscribeToValueChanges(undefined, extra && extra.classValidatorOptions);
        classValidators = null;
        formGroup = null;
        if (extra.validateAllFormFields) {
            dynamicFormGroup.validate();
        }
        // console.timeEnd(factoryModel.toString());
        return dynamicFormGroup;
    }
    /**
     * @protected
     * @return {?}
     */
    factoryFormBuilder() {
        return new FormBuilder();
    }
    /**
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} fields
     * @param {?=} validatorOrOpts
     * @param {?=} asyncValidator
     * @param {?=} options
     * @return {?}
     */
    factoryDynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options) {
        var _a, _b, _c;
        /** @type {?} */
        const formGroup = ((_a = this.options) === null || _a === void 0 ? void 0 : _a.factoryDynamicFormGroup) ? this.options.factoryDynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options)
            : new DynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options);
        formGroup.dynamicFormBuilder = this;
        formGroup.originalFormBuilder = ((_b = this.options) === null || _b === void 0 ? void 0 : _b.factoryFormBuilder) ? (_c = this.options) === null || _c === void 0 ? void 0 : _c.factoryFormBuilder() : this.factoryFormBuilder();
        return formGroup;
    }
    // *******************
    // Helpers
    /**
     * Recursively creates an empty object from the data provided
     * @protected
     * @template TModel
     * @param {?} factoryModel
     * @param {?=} data
     * @return {?}
     */
    createEmptyObject(factoryModel, data = {}) {
        /** @type {?} */
        let modifed = false;
        /** @type {?} */
        let object = factoryModel ? this.emptyDynamicFormGroup.plainToClass(factoryModel, data) : data;
        /** @type {?} */
        let fields = Object.keys(object);
        /** @type {?} */
        let objectFieldNameLength;
        /** @type {?} */
        let objectFieldName0;
        fields.forEach((/**
         * @param {?} fieldName
         * @return {?}
         */
        (fieldName) => {
            /** @type {?} */
            const object1 = object[fieldName];
            objectFieldNameLength = object1 && Array.isArray(object1) ? object1.length : undefined;
            if (objectFieldNameLength !== undefined) {
                objectFieldName0 = object1[0];
                if (objectFieldNameLength === 1 && Object.keys(objectFieldName0).length > 0 && objectFieldName0.constructor) {
                    object[fieldName] = [this.createEmptyObject(objectFieldName0.constructor)];
                }
                if (objectFieldNameLength === 0) {
                    data[fieldName] = [{}];
                    modifed = true;
                }
            }
            else {
                data[fieldName] = undefined;
            }
        }));
        if (modifed) {
            object = null;
            fields = null;
            return this.createEmptyObject(factoryModel, data);
        }
        fields = null;
        return object;
    }
}
if (false) {
    /**
     * @type {?}
     * @protected
     */
    DynamicFormBuilder.prototype.emptyDynamicFormGroup;
    /**
     * @type {?}
     * @protected
     */
    DynamicFormBuilder.prototype.options;
}

/**
 * @fileoverview added by tsickle
 * Generated from: index.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: public_api.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: ngx-dynamic-form-builder.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { DEFAULT_CLASS_TRANSFORM_OPTIONS, DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS, DEFAULT_CLASS_VALIDATOR_OPTIONS, DynamicFormBuilder, DynamicFormControl, DynamicFormGroup, ExposeNested, FOREVER_INVALID_NAME, foreverInvalid, getClassValidators, getGlobal, getOrSetEmptyObject, getValidatorMessagesStorage, getValidatorTitlesStorage, hasToJSON, isAbstractControlOptions, isDynamicFormGroupConfig, isLegacyOrOpts, mergeErrors, transformValidationErrors, updateValidatorMessagesStorage, updateValidatorTitlesStorage };
//# sourceMappingURL=ngx-dynamic-form-builder.js.map
