import { AbstractControl, AbstractControlOptions, AsyncValidatorFn, FormBuilder, FormGroup, ValidatorFn } from '@angular/forms';
import { ClassTransformOptions } from 'class-transformer';
import { ClassConstructor } from 'class-transformer';
import { ValidatorOptions } from 'class-validator-multi-lang';
import 'reflect-metadata';
import { BehaviorSubject, Subject, Subscription } from 'rxjs';
import { Dictionary } from '../types/dictionary';
import { DynamicFormGroupConfig } from '../types/dynamic-form-group-config';
import { FormModel } from '../types/form-model';
import { ShortValidationErrors } from '../types/short-validation-errors';
import { DynamicFormBuilder } from './dynamic-form-builder';
export declare class DynamicFormGroup<TModel> extends FormGroup {
    factoryModel: ClassConstructor<TModel>;
    fields?: FormModel<TModel> | undefined;
    protected _validatorOrOpts?: ValidatorFn | ValidatorFn[] | AbstractControlOptions | null | undefined;
    protected _asyncValidator?: AsyncValidatorFn | AsyncValidatorFn[] | null | undefined;
    protected readonly options?: DynamicFormGroupConfig | undefined;
    nativeValidateErrors: BehaviorSubject<Dictionary<any>>;
    customValidateErrors: BehaviorSubject<ShortValidationErrors>;
    formErrors: ShortValidationErrors;
    formFields: Dictionary;
    objectChange: Subject<unknown>;
    valueChangesSubscription: Subscription;
    dynamicFormBuilder: DynamicFormBuilder;
    originalFormBuilder: FormBuilder;
    protected _object: TModel;
    protected _externalErrors: ShortValidationErrors;
    protected _validatorOptions: ValidatorOptions;
    protected _classTransformOptions: ClassTransformOptions;
    protected _classTransformToPlainOptions: ClassTransformOptions;
    protected _validateSubscription: Subscription | undefined;
    protected _validateAllFormFields: boolean;
    constructor(factoryModel: ClassConstructor<TModel>, fields?: FormModel<TModel> | undefined, _validatorOrOpts?: ValidatorFn | ValidatorFn[] | AbstractControlOptions | null | undefined, _asyncValidator?: AsyncValidatorFn | AsyncValidatorFn[] | null | undefined, options?: DynamicFormGroupConfig | undefined);
    subscribeToValueChanges(externalErrors?: ShortValidationErrors, validatorOptions?: ValidatorOptions): void;
    set externalErrors(externalErrors: ShortValidationErrors);
    get externalErrors(): ShortValidationErrors;
    set classTransformOptions(classTransformOptions: ClassTransformOptions);
    get classTransformOptions(): ClassTransformOptions;
    set classTransformToPlainOptions(classTransformToPlainOptions: ClassTransformOptions);
    get classTransformToPlainOptions(): ClassTransformOptions;
    set validatorOptions(validatorOptions: ValidatorOptions);
    get validatorOptions(): ValidatorOptions;
    set object(object: TModel);
    get object(): TModel;
    set json(jsonData: any);
    get json(): any;
    validateStream(externalErrors?: ShortValidationErrors, validatorOptions?: ValidatorOptions): import("rxjs").Observable<[{
        [key: string]: string;
    }, {
        [key: string]: string;
    } | undefined]>;
    validateAsync(externalErrors?: ShortValidationErrors, validatorOptions?: ValidatorOptions): Promise<[{
        [key: string]: string;
    }, {
        [key: string]: string;
    } | undefined]>;
    validate(externalErrors?: ShortValidationErrors, validatorOptions?: ValidatorOptions): void;
    setCustomErrors(allErrors: any, force?: boolean): void;
    protected collectErrors(control: Dictionary, isRoot?: boolean): any;
    validateAllFormFields(): void;
    resetValidateAllFormFields(): void;
    classToClass<TClassModel>(object: TClassModel): any;
    plainToClass<TClassModel, Object>(cls: ClassConstructor<TClassModel>, plain: Object): any;
    classToPlain<TClassModel>(object: TClassModel): any;
    /**
     * @deprecated please use sync method setExternalErrors(errors)
     */
    setExternalErrorsAsync(externalErrors: ShortValidationErrors): Promise<void>;
    setExternalErrors(externalErrors: ShortValidationErrors): void;
    getExternalErrors(): ShortValidationErrors;
    /**
     * @deprecated please use sync method setExternalErrors({})
     */
    clearExternalErrorsAsync(): Promise<void>;
    clearExternalErrors(): void;
    setValidatorOptions(validatorOptions?: ValidatorOptions): void;
    getValidatorOptions(): ValidatorOptions;
    protected onlyFields(fields?: FormModel<any>): Dictionary;
    protected markAsInvalidForExternalErrors(errors: ShortValidationErrors, controls?: Dictionary<AbstractControl>): void;
    /**
     * Recursively gets all values from the form controls and all sub form group and array controls and returns it as
     * an object
     */
    protected getObject(): TModel;
    /**
     * Sets the value of every control on the form and recursively sets the values of the controls
     * on all sub form groups
     *
     * @param object the data to assign to all controls of the form group and sub groups
     */
    protected setObject(object: TModel): void;
    protected getJSON(): any;
    protected setJSON(jsonData: any): void;
}
export declare function getClassValidators<TModel>(factoryModel: ClassConstructor<TModel>, fields?: Dictionary, validatorOptions?: ValidatorOptions): {};
